/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * textbox.cpp: 
 *
 * Contact:
 *   Moonlight List (moonlight-list@lists.ximian.com)
 *
 * Copyright 2007 Novell, Inc. (http://www.novell.com)
 *
 * See the LICENSE file included with the distribution for details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <cairo.h>

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "dependencyproperty.h"
#include "file-downloader.h"
#include "contentcontrol.h"
#include "timemanager.h"
#include "runtime.h"
#include "textbox.h"
#include "border.h"
#include "panel.h"
#include "utils.h"
#include "uri.h"


//
// TextBuffer
//

#define UNICODE_LEN(size) (sizeof (gunichar) * (size))
#define UNICODE_OFFSET(buf,offset) (((char *) buf) + UNICODE_LEN (offset))

class TextBuffer {
	int allocated;
	
	bool Resize (int needed)
	{
		int new_size = allocated;
		bool resize = false;
		void *buf;
		
		if (allocated >= needed + 128) {
			while (new_size >= needed + 128)
				new_size -= 128;
			resize = true;
		} else if (allocated < needed) {
			while (new_size < needed)
				new_size += 128;
			resize = true;
		}
		
		if (resize) {
			if (!(buf = g_try_realloc (text, UNICODE_LEN (new_size)))) {
				// if new_size is < allocated, then we can pretend like we succeeded
				return new_size < allocated;
			}
			
			text = (gunichar *) buf;
			allocated = new_size;
		}
		
		return true;
	}
	
 public:
	gunichar *text;
	int len;
	
	TextBuffer (gunichar *text, int len)
	{
		this->allocated = len + 1;
		this->text = text;
		this->len = len;
	}
	
	TextBuffer ()
	{
		text = NULL;
		Reset ();
	}
	
	void Reset ()
	{
		text = (gunichar *) g_realloc (text, UNICODE_LEN (128));
		allocated = 128;
		text[0] = '\0';
		len = 0;
	}
	
	void Print ()
	{
		printf ("TextBuffer::text = \"");
		
		for (int i = 0; i < len; i++) {
			switch (text[i]) {
			case '\r':
				fputs ("\\r", stdout);
				break;
			case '\n':
				fputs ("\\n", stdout);
				break;
			case '\0':
				fputs ("\\0", stdout);
				break;
			case '\\':
				fputc ('\\', stdout);
				// fall thru
			default:
				fputc ((char) text[i], stdout);
				break;
			}
		}
		
		printf ("\";\n");
	}
	
	void Append (gunichar c)
	{
		if (!Resize (len + 2))
			return;
		
		text[len++] = c;
		text[len] = 0;
	}
	
	void Append (const gunichar *str, int count)
	{
		if (!Resize (len + count + 1))
			return;
		
		memcpy (UNICODE_OFFSET (text, len), str, UNICODE_LEN (count));
		len += count;
		text[len] = 0;
	}
	
	void Cut (int start, int length)
	{
		char *dest, *src;
		int beyond;
		
		if (length == 0 || start >= len)
			return;
		
		if (start + length > len)
			length = len - start;
		
		src = UNICODE_OFFSET (text, start + length);
		dest = UNICODE_OFFSET (text, start);
		beyond = len - (start + length);
		
		memmove (dest, src, UNICODE_LEN (beyond + 1));
		len -= length;
	}
	
	void Insert (int index, gunichar c)
	{
		if (!Resize (len + 2))
			return;
		
		if (index < len) {
			// shift all chars beyond position @index down by 1 char
			memmove (UNICODE_OFFSET (text, index + 1), UNICODE_OFFSET (text, index), UNICODE_LEN ((len - index) + 1));
			text[index] = c;
			len++;
		} else {
			text[len++] = c;
			text[len] = 0;
		}
	}
	
	void Insert (int index, const gunichar *str, int count)
	{
		if (!Resize (len + count + 1))
			return;
		
		if (index < len) {
			// shift all chars beyond position @index down by @count chars
			memmove (UNICODE_OFFSET (text, index + count), UNICODE_OFFSET (text, index), UNICODE_LEN ((len - index) + 1));
			
			// insert @count chars of @str into our buffer at position @index
			memcpy (UNICODE_OFFSET (text, index), str, UNICODE_LEN (count));
			len += count;
		} else {
			// simply append @count chars of @str onto the end of our buffer
			memcpy (UNICODE_OFFSET (text, len), str, UNICODE_LEN (count));
			len += count;
			text[len] = 0;
		}
	}
	
	void Prepend (gunichar c)
	{
		if (!Resize (len + 2))
			return;
		
		// shift the entire buffer down by 1 char
		memmove (UNICODE_OFFSET (text, 1), text, UNICODE_LEN (len + 1));
		text[0] = c;
		len++;
	}
	
	void Prepend (const gunichar *str, int count)
	{
		if (!Resize (len + count + 1))
			return;
		
		// shift the endtire buffer down by @count chars
		memmove (UNICODE_OFFSET (text, count), text, UNICODE_LEN (len + 1));
		
		// copy @count chars of @str into the beginning of our buffer
		memcpy (text, str, UNICODE_LEN (count));
		len += count;
	}
	
	void Replace (int start, int length, const gunichar *str, int count)
	{
		char *dest, *src;
		int beyond;
		
		if (start > len)
			return;
		
		if (start + length > len)
			length = len - start;
		
		// Check for the easy cases first...
		if (length == 0) {
			Insert (start, str, count);
			return;
		} else if (count == 0) {
			Cut (start, length);
			return;
		} else if (count == length) {
			memcpy (UNICODE_OFFSET (text, start), str, UNICODE_LEN (count));
			return;
		}
		
		if (count > length && !Resize (len + (count - length) + 1))
			return;
		
		// calculate the number of chars beyond @start that won't be cut
		beyond = len - (start + length);
		
		// shift all chars beyond position (@start + length) into position...
		dest = UNICODE_OFFSET (text, start + count);
		src = UNICODE_OFFSET (text, start + length);
		memmove (dest, src, UNICODE_LEN (beyond + 1));
		
		// copy @count chars of @str into our buffer at position @start
		memcpy (UNICODE_OFFSET (text, start), str, UNICODE_LEN (count));
		
		len = (len - length) + count;
	}
	
	gunichar *Substring (int start, int length = -1)
	{
		gunichar *substr;
		size_t n_bytes;
		
		if (start < 0 || start > len || length == 0)
			return NULL;
		
		if (length < 0)
			length = len - start;
		
		n_bytes = sizeof (gunichar) * (length + 1);
		substr = (gunichar *) g_malloc (n_bytes);
		n_bytes -= sizeof (gunichar);
		
		memcpy (substr, text + start, n_bytes);
		substr[length] = 0;
		
		return substr;
	}
};


//
// TextBoxUndoActions
//

enum TextBoxUndoActionType {
	TextBoxUndoActionTypeInsert,
	TextBoxUndoActionTypeDelete,
	TextBoxUndoActionTypeReplace,
};

class TextBoxUndoAction : public List::Node {
 public:
	TextBoxUndoActionType type;
	int selection_anchor;
	int selection_cursor;
	int length;
	int start;
};

class TextBoxUndoActionInsert : public TextBoxUndoAction {
 public:
	TextBuffer *buffer;
	bool growable;
	
	TextBoxUndoActionInsert (int selection_anchor, int selection_cursor, int start, gunichar c);
	TextBoxUndoActionInsert (int selection_anchor, int selection_cursor, int start, gunichar *inserted, int length);
	virtual ~TextBoxUndoActionInsert ();
	
	bool Insert (int start, const gunichar *text, int len);
	bool Insert (int start, gunichar c);
};

class TextBoxUndoActionDelete : public TextBoxUndoAction {
 public:
	gunichar *text;
	
	TextBoxUndoActionDelete (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length);
	virtual ~TextBoxUndoActionDelete ();
};

class TextBoxUndoActionReplace : public TextBoxUndoAction {
 public:
	gunichar *inserted;
	gunichar *deleted;
	int inlen;
	
	TextBoxUndoActionReplace (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length, gunichar *inserted, int inlen);
	TextBoxUndoActionReplace (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length, gunichar c);
	virtual ~TextBoxUndoActionReplace ();
};

class TextBoxUndoStack {
	int max_count;
	List *list;
	
 public:
	TextBoxUndoStack (int max_count);
	~TextBoxUndoStack ();
	
	bool IsEmpty ();
	void Clear ();
	
	void Push (TextBoxUndoAction *action);
	TextBoxUndoAction *Peek ();
	TextBoxUndoAction *Pop ();
};

TextBoxUndoActionInsert::TextBoxUndoActionInsert (int selection_anchor, int selection_cursor, int start, gunichar c)
{
	this->type = TextBoxUndoActionTypeInsert;
	this->selection_anchor = selection_anchor;
	this->selection_cursor = selection_cursor;
	this->start = start;
	this->length = 1;
	
	this->buffer = new TextBuffer ();
	this->buffer->Append (c);
	this->growable = true;
}

TextBoxUndoActionInsert::TextBoxUndoActionInsert (int selection_anchor, int selection_cursor, int start, gunichar *inserted, int length)
{
	this->type = TextBoxUndoActionTypeInsert;
	this->selection_anchor = selection_anchor;
	this->selection_cursor = selection_cursor;
	this->length = length;
	this->start = start;
	
	this->buffer = new TextBuffer (inserted, length);
	this->growable = false;
}

TextBoxUndoActionInsert::~TextBoxUndoActionInsert ()
{
	delete buffer;
}

bool
TextBoxUndoActionInsert::Insert (int start, const gunichar *text, int len)
{
	if (!growable || start != (this->start + length))
		return false;
	
	buffer->Append (text, len);
	length += len;
	
	return true;
}

bool
TextBoxUndoActionInsert::Insert (int start, gunichar c)
{
	if (!growable || start != (this->start + length))
		return false;
	
	buffer->Append (c);
	length++;
	
	return true;
}

TextBoxUndoActionDelete::TextBoxUndoActionDelete (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length)
{
	this->type = TextBoxUndoActionTypeDelete;
	this->selection_anchor = selection_anchor;
	this->selection_cursor = selection_cursor;
	this->length = length;
	this->start = start;
	
	this->text = buffer->Substring (start, length);
}

TextBoxUndoActionDelete::~TextBoxUndoActionDelete ()
{
	g_free (text);
}

TextBoxUndoActionReplace::TextBoxUndoActionReplace (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length, gunichar *inserted, int inlen)
{
	this->type = TextBoxUndoActionTypeReplace;
	this->selection_anchor = selection_anchor;
	this->selection_cursor = selection_cursor;
	this->length = length;
	this->start = start;
	
	this->deleted = buffer->Substring (start, length);
	this->inserted = inserted;
	this->inlen = inlen;
}

TextBoxUndoActionReplace::TextBoxUndoActionReplace (int selection_anchor, int selection_cursor, TextBuffer *buffer, int start, int length, gunichar c)
{
	this->type = TextBoxUndoActionTypeReplace;
	this->selection_anchor = selection_anchor;
	this->selection_cursor = selection_cursor;
	this->length = length;
	this->start = start;
	
	this->deleted = buffer->Substring (start, length);
	this->inserted = g_new (gunichar, 2);
	memcpy (inserted, &c, sizeof (gunichar));
	inserted[1] = 0;
	this->inlen = 1;
}

TextBoxUndoActionReplace::~TextBoxUndoActionReplace ()
{
	g_free (inserted);
	g_free (deleted);
}


TextBoxUndoStack::TextBoxUndoStack (int max_count)
{
	this->max_count = max_count;
	this->list = new List ();
}

TextBoxUndoStack::~TextBoxUndoStack ()
{
	delete list;
}

bool
TextBoxUndoStack::IsEmpty ()
{
	return list->IsEmpty ();
}

void
TextBoxUndoStack::Clear ()
{
	list->Clear (true);
}

void
TextBoxUndoStack::Push (TextBoxUndoAction *action)
{
	if (list->Length () == max_count) {
		List::Node *node = list->Last ();
		list->Unlink (node);
		delete node;
	}
	
	list->Prepend (action);
}

TextBoxUndoAction *
TextBoxUndoStack::Pop ()
{
	List::Node *node = list->First ();
	
	if (node)
		list->Unlink (node);
	
	return (TextBoxUndoAction *) node;
}

TextBoxUndoAction *
TextBoxUndoStack::Peek ()
{
	return (TextBoxUndoAction *) list->First ();
}


//
// TextBoxBase
//

// emit state, also doubles as available event mask
#define NOTHING_CHANGED         (0)
#define SELECTION_CHANGED       (1 << 0)
#define TEXT_CHANGED            (1 << 1)

#define CONTROL_MASK GDK_CONTROL_MASK
#define SHIFT_MASK   GDK_SHIFT_MASK
#define ALT_MASK     GDK_MOD1_MASK

#define IsEOL(c) ((c) == '\r' || (c) == '\n')

static GdkWindow *
GetGdkWindow (TextBoxBase *textbox)
{
	MoonWindow *window;
	Surface *surface;
	
	if (!(surface = textbox->GetSurface ()))
		return NULL;
	
	if (!(window = surface->GetWindow ()))
		return NULL;
	
	return window->GetGdkWindow ();
}

static GtkClipboard *
GetClipboard (TextBoxBase *textbox, GdkAtom atom)
{
	GdkDisplay *display;
	GdkWindow *window;
	
	if (!(window = GetGdkWindow (textbox)))
		return NULL;
	
	if (!(display = gdk_drawable_get_display ((GdkDrawable *) window)))
		return NULL;
	
	return gtk_clipboard_get_for_display (display, atom);
}

void
TextBoxBase::Initialize (Type::Kind type, const char *type_name)
{
	ManagedTypeInfo *type_info = new ManagedTypeInfo ("System.Windows", type_name);
	
	SetObjectType (type);
	SetDefaultStyleKey (type_info);
	
	AddHandler (UIElement::MouseLeftButtonDownEvent, TextBoxBase::mouse_left_button_down, this);
	AddHandler (UIElement::MouseLeftButtonUpEvent, TextBoxBase::mouse_left_button_up, this);
	AddHandler (UIElement::MouseMoveEvent, TextBoxBase::mouse_move, this);
	AddHandler (UIElement::LostFocusEvent, TextBoxBase::focus_out, this);
	AddHandler (UIElement::GotFocusEvent, TextBoxBase::focus_in, this);
	AddHandler (UIElement::KeyDownEvent, TextBoxBase::key_down, this);
	AddHandler (UIElement::KeyUpEvent, TextBoxBase::key_up, this);
	
	font = new TextFontDescription ();
	font->SetFamily (GetFontFamily()->source);
	font->SetStretch (GetFontStretch());
	font->SetWeight (GetFontWeight());
	font->SetStyle (GetFontStyle());
	font->SetSize (GetFontSize());
	
	downloader = NULL;
	
	contentElement = NULL;
	
	im_ctx = gtk_im_multicontext_new ();
	gtk_im_context_set_use_preedit (im_ctx, false);
	
	g_signal_connect (im_ctx, "retrieve-surrounding", G_CALLBACK (TextBoxBase::retrieve_surrounding), this);
	g_signal_connect (im_ctx, "delete-surrounding", G_CALLBACK (TextBoxBase::delete_surrounding), this);
	g_signal_connect (im_ctx, "commit", G_CALLBACK (TextBoxBase::commit), this);
	
	undo = new TextBoxUndoStack (10);
	redo = new TextBoxUndoStack (10);
	buffer = new TextBuffer ();
	max_length = 0;
	
	emit = NOTHING_CHANGED;
	events_mask = 0;
	
	selection_anchor = 0;
	selection_cursor = 0;
	cursor_offset = 0.0;
	batch = 0;
	
	accepts_return = false;
	need_im_reset = false;
	is_read_only = false;
	have_offset = false;
	selecting = false;
	setvalue = true;
	captured = false;
	focused = false;
	view = NULL;
}

TextBoxBase::~TextBoxBase ()
{
	RemoveHandler (UIElement::MouseLeftButtonDownEvent, TextBoxBase::mouse_left_button_down, this);
	RemoveHandler (UIElement::MouseLeftButtonUpEvent, TextBoxBase::mouse_left_button_up, this);
	RemoveHandler (UIElement::MouseMoveEvent, TextBoxBase::mouse_move, this);
	RemoveHandler (UIElement::LostFocusEvent, TextBoxBase::focus_out, this);
	RemoveHandler (UIElement::GotFocusEvent, TextBoxBase::focus_in, this);
	RemoveHandler (UIElement::KeyDownEvent, TextBoxBase::key_down, this);
	RemoveHandler (UIElement::KeyUpEvent, TextBoxBase::key_up, this);
	
	ResetIMContext ();
	g_object_unref (im_ctx);
	
	CleanupDownloader ();
	
	delete buffer;
	delete undo;
	delete redo;
	delete font;
}

void
TextBoxBase::SetSurface (Surface *surface)
{
	Control::SetSurface (surface);
	
	gtk_im_context_set_client_window (im_ctx, GetGdkWindow (this));
}

void
TextBoxBase::CleanupDownloader ()
{
	if (downloader) {
		downloader->RemoveHandler (Downloader::CompletedEvent, downloader_complete, this);
		downloader->Abort ();
		downloader->unref ();
		downloader = NULL;
	}
}

double
TextBoxBase::GetCursorOffset ()
{
	if (!have_offset && view) {
		cursor_offset = view->GetCursor ().x;
		have_offset = true;
	}
	
	return cursor_offset;
}

int
TextBoxBase::CursorDown (int cursor, bool page)
{
	double y = view->GetCursor ().y;
	double x = GetCursorOffset ();
	TextLayoutLine *line;
	TextLayoutRun *run;
	int index, cur, n;
	guint i;
	
	if (!(line = view->GetLineFromY (y, &index)))
		return cursor;
	
	if (page) {
		// calculate the number of lines to skip over
		n = GetActualHeight () / line->height;
	} else {
		n = 1;
	}
	
	if (index + n >= view->GetLineCount ()) {
		// go to the end of the last line
		line = view->GetLineFromIndex (view->GetLineCount () - 1);
		
		for (cur = line->offset, i = 0; i < line->runs->len; i++) {
			run = (TextLayoutRun *) line->runs->pdata[i];
			cur += run->count;
		}
		
		have_offset = false;
		
		return cur;
	}
	
	line = view->GetLineFromIndex (index + n);
	
	return line->GetCursorFromX (Point (), x);
}

int
TextBoxBase::CursorUp (int cursor, bool page)
{
	double y = view->GetCursor ().y;
	double x = GetCursorOffset ();
	TextLayoutLine *line;
	int index, n;
	
	if (!(line = view->GetLineFromY (y, &index)))
		return cursor;
	
	if (page) {
		// calculate the number of lines to skip over
		n = GetActualHeight () / line->height;
	} else {
		n = 1;
	}
	
	if (index < n) {
		// go to the beginning of the first line
		have_offset = false;
		return 0;
	}
	
	line = view->GetLineFromIndex (index - n);
	
	return line->GetCursorFromX (Point (), x);
}

enum CharClass {
	CharClassUnknown,
	CharClassWhitespace,
	CharClassAlphaNumeric
};

static inline CharClass
char_class (gunichar c)
{
	if (g_unichar_isspace (c))
		return CharClassWhitespace;
	
	if (g_unichar_isalnum (c))
		return CharClassAlphaNumeric;
	
	return CharClassUnknown;
}

int
TextBoxBase::CursorNextWord (int cursor)
{
	int i, lf, cr;
	CharClass cc;
	
	// find the end of the current line
	cr = CursorLineEnd (cursor);
	if (buffer->text[cr] == '\r' && buffer->text[cr + 1] == '\n')
		lf = cr + 1;
	else
		lf = cr;
	
	// if the cursor is at the end of the line, return the starting offset of the next line
	if (cursor == cr || cursor == lf) {
		if (lf < buffer->len)
			return lf + 1;
		
		return cursor;
	}
	
	cc = char_class (buffer->text[cursor]);
	i = cursor;
	
	// skip over the word, punctuation, or run of whitespace
	while (i < cr && char_class (buffer->text[i]) == cc)
		i++;
	
	// skip any whitespace after the word/punct
	while (i < cr && char_class (buffer->text[i]) == CharClassWhitespace)
		i++;
	
	return i;
}

int
TextBoxBase::CursorPrevWord (int cursor)
{
	int begin, i, cr, lf;
	CharClass cc;
	
	// find the beginning of the current line
	lf = CursorLineBegin (cursor) - 1;
	
	if (lf > 0 && buffer->text[lf] == '\n' && buffer->text[lf - 1] == '\r')
		cr = lf - 1;
	else
		cr = lf;
	
	// if the cursor is at the beginning of the line, return the end of the prev line
	if (cursor - 1 == lf) {
		if (cr > 0)
			return cr;
		
		return 0;
	}
	
	cc = char_class (buffer->text[cursor - 1]);
	begin = lf + 1;
	i = cursor;
	
	// skip over the word, punctuation, or run of whitespace
	while (i > begin && char_class (buffer->text[i - 1]) == cc)
		i--;
	
	// if the cursor was at whitespace, skip back a word too
	if (cc == CharClassWhitespace && i > begin) {
		cc = char_class (buffer->text[i - 1]);
		while (i > begin && char_class (buffer->text[i - 1]) == cc)
			i--;
	}
	
	return i;
}

int
TextBoxBase::CursorLineBegin (int cursor)
{
	int cur = cursor;
	
	// find the beginning of the line
	while (cur > 0 && !IsEOL (buffer->text[cur - 1]))
		cur--;
	
	return cur;
}

int
TextBoxBase::CursorLineEnd (int cursor, bool include)
{
	int cur = cursor;
	
	// find the end of the line
	while (cur < buffer->len && !IsEOL (buffer->text[cur]))
		cur++;
	
	if (include && cur < buffer->len) {
		if (buffer->text[cur] == '\r' && buffer->text[cur + 1] == '\n')
			cur += 2;
		else
			cur++;
	}
	
	return cur;
}

void
TextBoxBase::KeyPressBackSpace (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	TextBoxUndoAction *action;
	int start = 0, length = 0;
	
	if ((modifiers & (ALT_MASK | SHIFT_MASK)) != 0)
		return;
	
	if (cursor != anchor) {
		// BackSpace w/ active selection: delete the selected text
		length = abs (cursor - anchor);
		start = MIN (anchor, cursor);
	} else if ((modifiers & CONTROL_MASK) != 0) {
		// Ctrl+BackSpace: delete the word ending at the cursor
		start = CursorPrevWord (cursor);
		length = cursor - start;
	} else if (cursor > 0) {
		// BackSpace: delete the char before the cursor position
		if (cursor >= 2 && buffer->text[cursor - 1] == '\n' && buffer->text[cursor - 2] == '\r') {
			start = cursor - 2;
			length = 2;
		} else {
			start = cursor - 1;
			length = 1;
		}
	}
	
	if (length > 0) {
		action = new TextBoxUndoActionDelete (selection_anchor, selection_cursor, buffer, start, length);
		undo->Push (action);
		redo->Clear ();
		
		buffer->Cut (start, length);
		emit |= TEXT_CHANGED;
		anchor = start;
		cursor = start;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::KeyPressDelete (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	TextBoxUndoAction *action;
	int start = 0, length = 0;
	
	if ((modifiers & (ALT_MASK | SHIFT_MASK)) != 0)
		return;
	
	if (cursor != anchor) {
		// Delete w/ active selection: delete the selected text
		length = abs (cursor - anchor);
		start = MIN (anchor, cursor);
	} else if ((modifiers & CONTROL_MASK) != 0) {
		// Ctrl+Delete: delete the word starting at the cursor
		length = CursorNextWord (cursor) - cursor;
		start = cursor;
	} else if (cursor < buffer->len) {
		// Delete: delete the char after the cursor position
		if (buffer->text[cursor] == '\r' && buffer->text[cursor + 1] == '\n')
			length = 2;
		else
			length = 1;
		
		start = cursor;
	}
	
	if (length > 0) {
		action = new TextBoxUndoActionDelete (selection_anchor, selection_cursor, buffer, start, length);
		undo->Push (action);
		redo->Clear ();
		
		buffer->Cut (start, length);
		emit |= TEXT_CHANGED;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::KeyPressPageDown (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	bool have;
	
	if ((modifiers & (CONTROL_MASK | ALT_MASK)) != 0)
		return;
	
	// move the cursor down one page from its current position
	cursor = CursorDown (cursor, true);
	have = have_offset;
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
		have_offset = have;
	}
}

void
TextBoxBase::KeyPressPageUp (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	bool have;
	
	if ((modifiers & (CONTROL_MASK | ALT_MASK)) != 0)
		return;
	
	// move the cursor up one page from its current position
	cursor = CursorUp (cursor, true);
	have = have_offset;
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
		have_offset = have;
	}
}

void
TextBoxBase::KeyPressDown (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	bool have;
	
	if ((modifiers & (CONTROL_MASK | ALT_MASK)) != 0)
		return;
	
	// move the cursor down by one line from its current position
	cursor = CursorDown (cursor, false);
	have = have_offset;
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
		have_offset = have;
	}
}

void
TextBoxBase::KeyPressUp (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	bool have;
	
	if ((modifiers & (CONTROL_MASK | ALT_MASK)) != 0)
		return;
	
	// move the cursor up by one line from its current position
	cursor = CursorUp (cursor, false);
	have = have_offset;
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
		have_offset = have;
	}
}

void
TextBoxBase::KeyPressHome (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	
	if ((modifiers & ALT_MASK) != 0)
		return;
	
	if ((modifiers & CONTROL_MASK) != 0) {
		// move the cursor to the beginning of the buffer
		cursor = 0;
	} else {
		// move the cursor to the beginning of the line
		cursor = CursorLineBegin (cursor);
	}
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
		have_offset = false;
	}
}

void
TextBoxBase::KeyPressEnd (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	
	if ((modifiers & ALT_MASK) != 0)
		return;
	
	if ((modifiers & CONTROL_MASK) != 0) {
		// move the cursor to the end of the buffer
		cursor = buffer->len;
	} else {
		// move the cursor to the end of the line
		cursor = CursorLineEnd (cursor);
	}
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::KeyPressRight (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	
	if ((modifiers & ALT_MASK) != 0)
		return;
	
	if ((modifiers & CONTROL_MASK) != 0) {
		// move the cursor to beginning of the next word
		cursor = CursorNextWord (cursor);
	} else {
		// move the cursor forward one character
		if (buffer->text[cursor] == '\r' && buffer->text[cursor + 1] == '\n') 
			cursor += 2;
		else if (cursor < buffer->len)
			cursor++;
	}
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::KeyPressLeft (GdkModifierType modifiers)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	
	if ((modifiers & ALT_MASK) != 0)
		return;
	
	if ((modifiers & CONTROL_MASK) != 0) {
		// move the cursor to the beginning of the previous word
		cursor = CursorPrevWord (cursor);
	} else {
		// move the cursor backward one character
		if (cursor >= 2 && buffer->text[cursor - 2] == '\r' && buffer->text[cursor - 1] == '\n')
			cursor -= 2;
		else if (cursor > 0)
			cursor--;
	}
	
	if ((modifiers & SHIFT_MASK) == 0) {
		// clobber the selection
		anchor = cursor;
	}
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::KeyPressUnichar (gunichar c)
{
	int length = abs (selection_cursor - selection_anchor);
	int start = MIN (selection_anchor, selection_cursor);
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	TextBoxUndoAction *action;
	
	if ((max_length > 0 && buffer->len >= max_length) || ((c == '\r') && !accepts_return))
		return;
	
	if (length > 0) {
		// replace the currently selected text
		action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, start, length, c);
		undo->Push (action);
		redo->Clear ();
		
		buffer->Replace (start, length, &c, 1);
	} else {
		// insert the text at the cursor position
		TextBoxUndoActionInsert *insert = NULL;
		
		if ((action = undo->Peek ()) && action->type == TextBoxUndoActionTypeInsert) {
			insert = (TextBoxUndoActionInsert *) action;
			
			if (!insert->Insert (start, c))
				insert = NULL;
		}
		
		if (!insert) {
			insert = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, start, c);
			undo->Push (insert);
		}
		
		redo->Clear ();
		
		buffer->Insert (start, c);
	}
	
	emit |= TEXT_CHANGED;
	cursor = start + 1;
	anchor = cursor;
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
}

void
TextBoxBase::BatchPush ()
{
	batch++;
}

void
TextBoxBase::BatchPop ()
{
	if (batch == 0) {
		g_warning ("TextBoxBase batch underflow");
		return;
	}
	
	batch--;
}

void
TextBoxBase::emit_selection_changed (EventObject *sender)
{
	((TextBoxBase *) sender)->EmitSelectionChanged ();
}

void
TextBoxBase::EmitSelectionChangedAsync ()
{
	if (events_mask & SELECTION_CHANGED)
		AddTickCall (TextBoxBase::emit_selection_changed);
	
	emit &= ~SELECTION_CHANGED;
}

void
TextBoxBase::emit_text_changed (EventObject *sender)
{
	((TextBoxBase *) sender)->EmitTextChanged ();
}

void
TextBoxBase::EmitTextChangedAsync ()
{
	if (events_mask & TEXT_CHANGED)
		AddTickCall (TextBoxBase::emit_text_changed);
	
	emit &= ~TEXT_CHANGED;
}

void
TextBoxBase::SyncAndEmit (bool sync_text)
{
	if (batch != 0 || emit == NOTHING_CHANGED)
		return;
	
	if (sync_text && (emit & TEXT_CHANGED))
		SyncText ();
	
	if (emit & SELECTION_CHANGED)
		SyncSelectedText ();
	
	if (emit & TEXT_CHANGED)
		EmitTextChangedAsync ();
	
	if (emit & SELECTION_CHANGED)
		EmitSelectionChangedAsync ();
	
	emit = NOTHING_CHANGED;
}

void
TextBoxBase::Paste (GtkClipboard *clipboard, const char *str)
{
	TextBoxUndoAction *action;
	int start, length;
	gunichar *text;
	glong len;
	
	length = abs (selection_cursor - selection_anchor);
	start = MIN (selection_anchor, selection_cursor);
	
	if (!(text = g_utf8_to_ucs4_fast (str ? str : "", -1, &len)))
		return;
	
	ResetIMContext ();
	
	if (length > 0) {
		// replace the currently selected text
		action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, start, length, text, len);
		
		buffer->Replace (start, length, text, len);
	} else {
		// insert the text at the cursor position
		action = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, start, text, len);
		
		buffer->Insert (start, text, len);
	}
	
	undo->Push (action);
	redo->Clear ();
	
	emit |= TEXT_CHANGED;
	start += len;
	
	BatchPush ();
	SetSelectionStart (start);
	SetSelectionLength (0);
	BatchPop ();
	
	SyncAndEmit ();
}

void
TextBoxBase::paste (GtkClipboard *clipboard, const char *text, gpointer closure)
{
	((TextBoxBase *) closure)->Paste (clipboard, text);
}

void
TextBoxBase::OnKeyDown (KeyEventArgs *args)
{
	GdkModifierType modifiers = (GdkModifierType) args->GetModifiers ();
	guint key = args->GetKeyVal ();
	GtkClipboard *clipboard;
	gunichar c;
	
	if (!is_read_only && gtk_im_context_filter_keypress (im_ctx, args->GetEvent ())) {
		args->SetHandled (true);
		need_im_reset = true;
		return;
	}
	
	ResetIMContext ();
	
	if (args->IsModifier ())
		return;
	
	// set 'emit' to NOTHING_CHANGED so that we can figure out
	// what has chanegd after applying the changes that this
	// keypress will cause.
	emit = NOTHING_CHANGED;
	BatchPush ();
	
	switch (key) {
	case GDK_Return:
		if (!is_read_only) {
			args->SetHandled (true);
			KeyPressUnichar ('\r');
		}
		break;
	case GDK_BackSpace:
		if (!is_read_only) {
			KeyPressBackSpace (modifiers);
			args->SetHandled (true);
		}
		break;
	case GDK_Delete:
		if (!is_read_only) {
			KeyPressDelete (modifiers);
			args->SetHandled (true);
		}
		break;
	case GDK_KP_Page_Down:
	case GDK_Page_Down:
		KeyPressPageDown (modifiers);
		break;
	case GDK_KP_Page_Up:
	case GDK_Page_Up:
		KeyPressPageUp (modifiers);
		break;
	case GDK_KP_Home:
	case GDK_Home:
		KeyPressHome (modifiers);
		break;
	case GDK_KP_End:
	case GDK_End:
		KeyPressEnd (modifiers);
		break;
	case GDK_KP_Right:
	case GDK_Right:
		KeyPressRight (modifiers);
		break;
	case GDK_KP_Left:
	case GDK_Left:
		KeyPressLeft (modifiers);
		break;
	case GDK_KP_Down:
	case GDK_Down:
		KeyPressDown (modifiers);
		break;
	case GDK_KP_Up:
	case GDK_Up:
		KeyPressUp (modifiers);
		break;
	default:
		if ((modifiers & (CONTROL_MASK | ALT_MASK | SHIFT_MASK)) == CONTROL_MASK) {
			switch (key) {
			case GDK_A:
			case GDK_a:
				// Ctrl+A => Select All
				args->SetHandled (true);
				SelectAll ();
				break;
			case GDK_C:
			case GDK_c:
				// Ctrl+C => Copy
				if ((clipboard = GetClipboard (this, GDK_SELECTION_CLIPBOARD))) {
					// copy selection to the clipboard
					gtk_clipboard_set_text (clipboard, GetSelectedText (), -1);
					args->SetHandled (true);
				}
				break;
			case GDK_X:
			case GDK_x:
				// Ctrl+X => Cut
				if ((clipboard = GetClipboard (this, GDK_SELECTION_CLIPBOARD))) {
					// copy selection to the clipboard and then cut
					gtk_clipboard_set_text (clipboard, GetSelectedText (), -1);
					if (!is_read_only)
						SetSelectedText ("");
					args->SetHandled (true);
				}
				break;
			case GDK_V:
			case GDK_v:
				// Ctrl+V => Paste
				if (!is_read_only && (clipboard = GetClipboard (this, GDK_SELECTION_CLIPBOARD))) {
					// paste clipboard contents to the buffer
					gtk_clipboard_request_text (clipboard, TextBoxBase::paste, this);
					args->SetHandled (true);
				}
				break;
			case GDK_Y:
			case GDK_y:
				// Ctrl+Y => Redo
				if (!is_read_only) {
					args->SetHandled (true);
					Redo ();
				}
				break;
			case GDK_Z:
			case GDK_z:
				// Ctrl+Z => Undo
				if (!is_read_only) {
					args->SetHandled (true);
					Undo ();
				}
				break;
			default:
				// unhandled Control commands
				break;
			}
		} else if ((modifiers & (CONTROL_MASK | ALT_MASK)) == 0) {
			// normal character input
			if ((c = args->GetUnicode ()) && !is_read_only) {
				args->SetHandled (true);
				KeyPressUnichar (c);
			}
		}
		break;
	}
	
	BatchPop ();
	
	SyncAndEmit ();
}

void
TextBoxBase::key_down (EventObject *sender, EventArgs *args, void *closure)
{
	((TextBoxBase *) closure)->OnKeyDown ((KeyEventArgs *) args);
}

void
TextBoxBase::OnKeyUp (KeyEventArgs *args)
{
	if (!is_read_only) {
		if (gtk_im_context_filter_keypress (im_ctx, args->GetEvent ()))
			need_im_reset = true;
		
		args->SetHandled (true);
	}
}

void
TextBoxBase::key_up (EventObject *sender, EventArgs *args, void *closure)
{
	((TextBoxBase *) closure)->OnKeyUp ((KeyEventArgs *) args);
}

bool
TextBoxBase::DeleteSurrounding (int offset, int n_chars)
{
	const char *delete_start, *delete_end;
	const char *text = GetActualText ();
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	TextBoxUndoAction *action;
	int start, length;
	
	if (is_read_only)
		return true;
	
	// get the utf-8 pointers so that we can use them to get gunichar offsets
	delete_start = g_utf8_offset_to_pointer (text, selection_cursor) + offset;
	delete_end = delete_start + n_chars;
	
	// get the character length/start index
	length = g_utf8_pointer_to_offset (delete_start, delete_end);
	start = g_utf8_pointer_to_offset (text, delete_start);
	
	if (length > 0) {
		action = new TextBoxUndoActionDelete (selection_anchor, selection_cursor, buffer, start, length);
		undo->Push (action);
		redo->Clear ();
		
		buffer->Cut (start, length);
		emit |= TEXT_CHANGED;
		anchor = start;
		cursor = start;
	}
	
	BatchPush ();
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
	
	BatchPop ();
	
	SyncAndEmit ();
	
	return true;
}

gboolean
TextBoxBase::delete_surrounding (GtkIMContext *context, int offset, int n_chars, gpointer user_data)
{
	return ((TextBoxBase *) user_data)->DeleteSurrounding (offset, n_chars);
}

bool
TextBoxBase::RetrieveSurrounding ()
{
	const char *text = GetActualText ();
	const char *cursor = g_utf8_offset_to_pointer (text, selection_cursor);
	
	gtk_im_context_set_surrounding (im_ctx, text, -1, cursor - text);
	
	return true;
}

gboolean
TextBoxBase::retrieve_surrounding (GtkIMContext *context, gpointer user_data)
{
	return ((TextBoxBase *) user_data)->RetrieveSurrounding ();
}

void
TextBoxBase::Commit (const char *str)
{
	TextBoxUndoAction *action;
	int anchor, cursor;
	int start, length;
	gunichar *text;
	glong len;
	
	if (is_read_only)
		return;
	
	length = abs (selection_cursor - selection_anchor);
	start = MIN (selection_anchor, selection_cursor);
	
	if (!(text = g_utf8_to_ucs4_fast (str ? str : "", -1, &len)))
		return;
	
	if (length > 0) {
		// replace the currently selected text
		action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, start, length, text, len);
		undo->Push (action);
		redo->Clear ();
		
		buffer->Replace (start, length, text, len);
	} else {
		// insert the text at the cursor position
		TextBoxUndoActionInsert *insert = NULL;
		
		if ((action = undo->Peek ()) && action->type == TextBoxUndoActionTypeInsert) {
			insert = (TextBoxUndoActionInsert *) action;
			
			if (!insert->Insert (start, text, len))
				insert = NULL;
		}
		
		if (!insert) {
			insert = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, start, text, len);
			undo->Push (insert);
		}
		
		redo->Clear ();
		
		buffer->Insert (start, text, len);
	}
	
	emit = TEXT_CHANGED;
	cursor = start + len;
	anchor = cursor;
	
	BatchPush ();
	
	// check to see if selection has changed
	if (selection_anchor != anchor || selection_cursor != cursor) {
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		emit |= SELECTION_CHANGED;
	}
	
	BatchPop ();
	
	SyncAndEmit ();
}

void
TextBoxBase::commit (GtkIMContext *context, const char *str, gpointer user_data)
{
	((TextBoxBase *) user_data)->Commit (str);
}

void
TextBoxBase::ResetIMContext ()
{
	if (need_im_reset) {
		gtk_im_context_reset (im_ctx);
		need_im_reset = false;
	}
}

void
TextBoxBase::OnMouseLeftButtonDown (MouseEventArgs *args)
{
	GdkEventButton *event = (GdkEventButton *) args->GetEvent ();
	int cursor, start, end;
	double x, y;
	
	args->SetHandled (true);
	Focus ();
	
	if (view) {
		args->GetPosition (view, &x, &y);
		
		cursor = view->GetCursorFromXY (x, y);
		
		ResetIMContext ();
		
		switch (event->type) {
		case GDK_3BUTTON_PRESS:
			// Note: Silverlight doesn't implement this, but to
			// be consistent with other TextEntry-type
			// widgets, we will.
			//
			// Triple-Click: select the line
			if (captured)
				ReleaseMouseCapture ();
			start = CursorLineBegin (cursor);
			end = CursorLineEnd (cursor, true);
			selecting = false;
			captured = false;
			break;
		case GDK_2BUTTON_PRESS:
			// Double-Click: select the word
			if (captured)
				ReleaseMouseCapture ();
			start = CursorPrevWord (cursor);
			end = CursorNextWord (cursor);
			selecting = false;
			captured = false;
			break;
		case GDK_BUTTON_PRESS:
		default:
			// Single-Click: cursor placement
			captured = CaptureMouse ();
			start = end = cursor;
			selecting = true;
			break;
		}
		
		BatchPush ();
		emit = NOTHING_CHANGED;
		SetSelectionLength (end - start);
		SetSelectionStart (start);
		BatchPop ();
		
		SyncAndEmit ();
	}
}

void
TextBoxBase::mouse_left_button_down (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxBase *) closure)->OnMouseLeftButtonDown ((MouseEventArgs *) args);
}

void
TextBoxBase::OnMouseLeftButtonUp (MouseEventArgs *args)
{
	if (captured)
		ReleaseMouseCapture ();
	
	args->SetHandled (true);
	selecting = false;
	captured = false;
}

void
TextBoxBase::mouse_left_button_up (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxBase *) closure)->OnMouseLeftButtonUp ((MouseEventArgs *) args);
}

void
TextBoxBase::OnMouseMove (MouseEventArgs *args)
{
	int anchor = selection_anchor;
	int cursor = selection_cursor;
	double x, y;
	
	if (selecting) {
		args->GetPosition (view, &x, &y);
		args->SetHandled (true);
		
		cursor = view->GetCursorFromXY (x, y);
		
		BatchPush ();
		emit = NOTHING_CHANGED;
		SetSelectionLength (abs (cursor - anchor));
		SetSelectionStart (MIN (anchor, cursor));
		selection_anchor = anchor;
		selection_cursor = cursor;
		BatchPop ();
		
		SyncAndEmit ();
	}
}

void
TextBoxBase::mouse_move (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxBase *) closure)->OnMouseMove ((MouseEventArgs *) args);
}

void
TextBoxBase::OnFocusOut (EventArgs *args)
{
	focused = false;
	
	if (view)
		view->OnFocusOut ();
	
	if (!is_read_only) {
		gtk_im_context_focus_out (im_ctx);
		need_im_reset = true;
	}
}

void
TextBoxBase::focus_out (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxBase *) closure)->OnFocusOut (args);
}

void
TextBoxBase::OnFocusIn (EventArgs *args)
{
	focused = true;
	
	if (view)
		view->OnFocusIn ();
	
	if (!is_read_only) {
		gtk_im_context_focus_in (im_ctx);
		need_im_reset = true;
	}
}

void
TextBoxBase::focus_in (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxBase *) closure)->OnFocusIn (args);
}

void
TextBoxBase::EmitCursorPositionChanged (double height, double x, double y)
{
	Emit (TextBoxBase::CursorPositionChangedEvent, new CursorPositionChangedEventArgs (height, x, y));
}

void
TextBoxBase::DownloaderComplete ()
{
	const char *path, *guid = NULL;
	InternalDownloader *idl;
	char *filename;
	size_t len;
	Uri *uri;
	
	// get the downloaded file path (enforces a mozilla workaround for files smaller than 64k)
	if (!(filename = downloader->GetDownloadedFilename (NULL)))
		return;
	
	g_free (filename);
	
	if (!(idl = downloader->GetInternalDownloader ()))
		return;
	
	if (!(idl->GetObjectType () == Type::FILEDOWNLOADER))
		return;
	
	uri = downloader->GetUri ();
	path = uri->GetPath ();
	
	len = path ? strlen (path) : 0;
	
	if (len > 6 && !g_ascii_strcasecmp (path + len - 6, ".odttf")) {
		// if the font file is obfuscated, use the basename of the path as the guid
		if (!(guid = strrchr (path, '/')))
			guid = path;
		else
			guid++;
	}
	
	// If the downloaded file was a zip file, this'll get the path to the
	// extracted zip directory, else it will simply be the path to the
	// downloaded file.
	if (!(path = ((FileDownloader *) idl)->GetUnzippedPath ()))
		return;
	
	font->SetFilename (path, guid);
	
	Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (TextBoxModelChangedFont, NULL));
}

void
TextBoxBase::downloader_complete (EventObject *sender, EventArgs *calldata, gpointer closure)
{
	((TextBoxBase *) closure)->DownloaderComplete ();
}

void
TextBoxBase::SetFontSource (Downloader *downloader)
{
	if (downloader) {
		this->downloader = downloader;
		downloader->ref ();
		
		downloader->AddHandler (downloader->CompletedEvent, downloader_complete, this);
		if (downloader->Started () || downloader->Completed ()) {
			if (downloader->Completed ())
				DownloaderComplete ();
		} else {
			// This is what actually triggers the download
			downloader->Send ();
		}
	} else {
		font->SetFilename (NULL, NULL);
	}
}

void
TextBoxBase::SetFontResource (const char *resource)
{
	Application *application = Application::GetCurrent ();
	const char *guid = NULL;
	Surface *surface;
	char *filename;
	size_t len;
	Uri *uri;
	
	ClearFontSource ();
	
	uri = new Uri ();
	
	if (!application || !uri->Parse (resource) || !(filename = application->GetResourceAsPath (uri))) {
		if ((surface = GetSurface ()) && (downloader = surface->CreateDownloader ())) {
			downloader->Open ("GET", resource, XamlPolicy);
			SetFontSource (downloader);
			downloader->unref ();
		} else {
			font->SetFilename (NULL, NULL);
		}
		
		delete uri;
		
		return;
	}
	
	delete uri;
	
	// check if the resource is an obfuscated font
	len = strlen (resource);
	if (len > 6 && !g_ascii_strcasecmp (resource + len - 6, ".odttf"))
		guid = resource;
	
	font->SetFilename (filename, guid);
	
	g_free (filename);
}

void
TextBoxBase::OnPropertyChanged (PropertyChangedEventArgs *args, MoonError *error)
{
	TextBoxModelChangeType changed = TextBoxModelChangedNothing;
	
	if (args->GetId () == Control::FontFamilyProperty) {
		FontFamily *family = args->GetNewValue() ? args->GetNewValue()->AsFontFamily () : NULL;
		const char *family_name = NULL;
		char *resource;
		
		if (family && family->source && (family_name = strchr (family->source, '#'))) {
			// FontFamily can reference a font resource in the form "<resource>#<family name>"
			resource = g_strndup (family->source, family_name - family->source);
			SetFontResource (resource);
			g_free (resource);
			family_name++;
		} else if (family) {
			family_name = family->source;
		}
		
		changed = TextBoxModelChangedFont;
		font->SetFamily (family_name);
	} else if (args->GetId () == Control::FontSizeProperty) {
		double size = args->GetNewValue()->AsDouble ();
		changed = TextBoxModelChangedFont;
		font->SetSize (size);
	} else if (args->GetId () == Control::FontStretchProperty) {
		FontStretches stretch = (FontStretches) args->GetNewValue()->AsInt32 ();
		changed = TextBoxModelChangedFont;
		font->SetStretch (stretch);
	} else if (args->GetId () == Control::FontStyleProperty) {
		FontStyles style = (FontStyles) args->GetNewValue()->AsInt32 ();
		changed = TextBoxModelChangedFont;
		font->SetStyle (style);
	} else if (args->GetId () == Control::FontWeightProperty) {
		FontWeights weight = (FontWeights) args->GetNewValue()->AsInt32 ();
		changed = TextBoxModelChangedFont;
		font->SetWeight (weight);
	}
	
	if (changed != TextBoxModelChangedNothing)
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (changed, args));
	
	if (args->GetProperty ()->GetOwnerType () != Type::TEXTBOXBASE) {
		Control::OnPropertyChanged (args, error);
		return;
	}
	
	NotifyListenersOfPropertyChange (args, error);
}

void
TextBoxBase::OnSubPropertyChanged (DependencyProperty *prop, DependencyObject *obj, PropertyChangedEventArgs *subobj_args)
{
	if (prop && (prop->GetId () == Control::BackgroundProperty ||
		     prop->GetId () == Control::ForegroundProperty)) {
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (TextBoxModelChangedBrush));
		Invalidate ();
	}
	
	if (prop->GetOwnerType () != Type::TEXTBOXBASE)
		Control::OnSubPropertyChanged (prop, obj, subobj_args);
}

void
TextBoxBase::OnApplyTemplate ()
{
	contentElement = GetTemplateChild ("ContentElement");
	
	if (contentElement == NULL) {
		g_warning ("TextBoxBase::OnApplyTemplate: no ContentElement found");
		Control::OnApplyTemplate ();
		return;
	}
	
	view = new TextBoxView ();
	view->SetTextBox (this);
	
	// Insert our TextBoxView
	if (contentElement->Is (Type::CONTENTCONTROL)) {
		ContentControl *control = (ContentControl *) contentElement;
		
		control->SetValue (ContentControl::ContentProperty, Value (view));
	} else if (contentElement->Is (Type::BORDER)) {
		Border *border = (Border *) contentElement;
		
		border->SetValue (Border::ChildProperty, Value (view));
	} else if (contentElement->Is (Type::PANEL)) {
		DependencyObjectCollection *children = ((Panel *) contentElement)->GetChildren ();
		
		children->Add (view);
	} else {
		g_warning ("TextBoxBase::OnApplyTemplate: don't know how to handle a ContentELement of type %s",
			   contentElement->GetType ()->GetName ());
		view->unref ();
	}
	
	Control::OnApplyTemplate ();
}

void
TextBoxBase::ClearSelection (int start)
{
	BatchPush ();
	SetSelectionStart (start);
	SetSelectionLength (0);
	BatchPop ();
}

void
TextBoxBase::Select (int start, int length)
{
	if ((start < 0) || (length < 0))
		return;
	
	if (start > buffer->len)
		start = buffer->len;
	
	if (length > (buffer->len - start))
		length = (buffer->len - start);
	
	BatchPush ();
	SetSelectionStart (start);
	SetSelectionLength (length);
	BatchPop ();
	
	ResetIMContext ();
	
	SyncAndEmit ();
}

void
TextBoxBase::SelectAll ()
{
	Select (0, buffer->len);
}

bool
TextBoxBase::CanUndo ()
{
	return !undo->IsEmpty ();
}

bool
TextBoxBase::CanRedo ()
{
	return !redo->IsEmpty ();
}

void
TextBoxBase::Undo ()
{
	TextBoxUndoActionReplace *replace;
	TextBoxUndoActionInsert *insert;
	TextBoxUndoActionDelete *dele;
	TextBoxUndoAction *action;
	int anchor, cursor;
	
	if (undo->IsEmpty ())
		return;
	
	action = undo->Pop ();
	redo->Push (action);
	
	switch (action->type) {
	case TextBoxUndoActionTypeInsert:
		insert = (TextBoxUndoActionInsert *) action;
		
		buffer->Cut (insert->start, insert->length);
		anchor = action->selection_anchor;
		cursor = action->selection_cursor;
		break;
	case TextBoxUndoActionTypeDelete:
		dele = (TextBoxUndoActionDelete *) action;
		
		buffer->Insert (dele->start, dele->text, dele->length);
		anchor = action->selection_anchor;
		cursor = action->selection_cursor;
		break;
	case TextBoxUndoActionTypeReplace:
		replace = (TextBoxUndoActionReplace *) action;
		
		buffer->Cut (replace->start, replace->inlen);
		buffer->Insert (replace->start, replace->deleted, replace->length);
		anchor = action->selection_anchor;
		cursor = action->selection_cursor;
		break;
	}
	
	BatchPush ();
	SetSelectionLength (abs (cursor - anchor));
	SetSelectionStart (MIN (anchor, cursor));
	emit = TEXT_CHANGED | SELECTION_CHANGED;
	selection_anchor = anchor;
	selection_cursor = cursor;
	BatchPop ();
	
	SyncAndEmit ();
}

void
TextBoxBase::Redo ()
{
	TextBoxUndoActionReplace *replace;
	TextBoxUndoActionInsert *insert;
	TextBoxUndoActionDelete *dele;
	TextBoxUndoAction *action;
	int anchor, cursor;
	
	if (redo->IsEmpty ())
		return;
	
	action = redo->Pop ();
	undo->Push (action);
	
	switch (action->type) {
	case TextBoxUndoActionTypeInsert:
		insert = (TextBoxUndoActionInsert *) action;
		
		buffer->Insert (insert->start, insert->buffer->text, insert->buffer->len);
		anchor = cursor = insert->start + insert->buffer->len;
		break;
	case TextBoxUndoActionTypeDelete:
		dele = (TextBoxUndoActionDelete *) action;
		
		buffer->Cut (dele->start, dele->length);
		anchor = cursor = dele->start;
		break;
	case TextBoxUndoActionTypeReplace:
		replace = (TextBoxUndoActionReplace *) action;
		
		buffer->Cut (replace->start, replace->length);
		buffer->Insert (replace->start, replace->inserted, replace->inlen);
		anchor = cursor = replace->start + replace->inlen;
		break;
	}
	
	BatchPush ();
	SetSelectionLength (abs (cursor - anchor));
	SetSelectionStart (MIN (anchor, cursor));
	emit = TEXT_CHANGED | SELECTION_CHANGED;
	selection_anchor = anchor;
	selection_cursor = cursor;
	BatchPop ();
	
	SyncAndEmit ();
}


//
// TextBoxDynamicPropertyValueProvider
//

class TextBoxDynamicPropertyValueProvider : public PropertyValueProvider {
	Value *selection_background;
	Value *selection_foreground;
	
 public:
	TextBoxDynamicPropertyValueProvider (DependencyObject *obj, PropertyPrecedence precedence) : PropertyValueProvider (obj, precedence)
	{
		selection_background = NULL;
		selection_foreground = NULL;
	}
	
	virtual ~TextBoxDynamicPropertyValueProvider ()
	{
		delete selection_background;
		delete selection_foreground;
	}
	
	virtual Value *GetPropertyValue (DependencyProperty *property)
	{
		if (property->GetId () == TextBox::SelectionBackgroundProperty) {
			return selection_background;
		} else if (property->GetId () == TextBox::SelectionForegroundProperty) {
			return selection_foreground;
		}
		
		return NULL;
	}
	
	void InitializeSelectionBrushes ()
	{
		if (!selection_background)
			selection_background = Value::CreateUnrefPtr (new SolidColorBrush ("#FF444444"));
		
		if (!selection_foreground)
			selection_foreground = Value::CreateUnrefPtr (new SolidColorBrush ("#FFFFFFFF"));
	}
};


//
// TextBox
//

TextBox::TextBox ()
{
	providers[PropertyPrecedence_DynamicValue] = new TextBoxDynamicPropertyValueProvider (this, PropertyPrecedence_DynamicValue);
	
	Initialize (Type::TEXTBOX, "System.Windows.Controls.TextBox");
	events_mask = TEXT_CHANGED | SELECTION_CHANGED;
}

void
TextBox::EmitSelectionChanged ()
{
	Emit (TextBox::SelectionChangedEvent, new RoutedEventArgs ());
}

void
TextBox::EmitTextChanged ()
{
	Emit (TextBox::TextChangedEvent, new TextChangedEventArgs ());
}

void
TextBox::SyncSelectedText ()
{
	if (selection_cursor != selection_anchor) {
		int length = abs (selection_cursor - selection_anchor);
		int start = MIN (selection_anchor, selection_cursor);
		char *text;
		
		text = g_ucs4_to_utf8 (buffer->text + start, length, NULL, NULL, NULL);
		
		setvalue = false;
		SetValue (TextBox::SelectedTextProperty, Value (text, true));
		setvalue = true;
	} else {
		setvalue = false;
		SetValue (TextBox::SelectedTextProperty, Value (""));
		setvalue = true;
	}
}

void
TextBox::SyncText ()
{
	char *text = g_ucs4_to_utf8 (buffer->text, buffer->len, NULL, NULL, NULL);
	
	setvalue = false;
	SetValue (TextBox::TextProperty, Value (text, true));
	setvalue = true;
}

void
TextBox::ClearFontSource ()
{
	ClearValue (TextBox::FontSourceProperty);
}

void
TextBox::OnPropertyChanged (PropertyChangedEventArgs *args, MoonError *error)
{
	TextBoxModelChangeType changed = TextBoxModelChangedNothing;
	DependencyProperty *prop;
	int start, length;
	
	if (args->GetId () == TextBox::AcceptsReturnProperty) {
		// update accepts_return state
		accepts_return = args->GetNewValue ()->AsBool ();
	} else if (args->GetId () == TextBox::IsReadOnlyProperty) {
		// update is_read_only state
		is_read_only = args->GetNewValue ()->AsBool ();
		
		if (focused) {
			if (is_read_only) {
				ResetIMContext ();
				gtk_im_context_focus_out (im_ctx);
			} else {
				gtk_im_context_focus_in (im_ctx);
			}
		}
	} else if (args->GetId () == TextBox::MaxLengthProperty) {
		// update max_length state
		max_length = args->GetNewValue ()->AsInt32 ();
	} else if (args->GetId () == TextBox::SelectedTextProperty) {
		if (setvalue) {
			Value *value = args->GetNewValue ();
			const char *str = value && value->AsString () ? value->AsString () : "";
			TextBoxUndoAction *action;
			gunichar *text;
			glong textlen;
			
			length = abs (selection_cursor - selection_anchor);
			start = MIN (selection_anchor, selection_cursor);
			
			if ((text = g_utf8_to_ucs4_fast (str, -1, &textlen))) {
				if (length > 0) {
					// replace the currently selected text
					action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, start, length, text, textlen);
					
					buffer->Replace (start, length, text, textlen);
				} else {
					// insert the text at the cursor
					action = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, start, text, textlen);
					
					buffer->Insert (start, text, textlen);
				}
				
				undo->Push (action);
				redo->Clear ();
				
				emit |= TEXT_CHANGED;
				
				ClearSelection (start + textlen);
				ResetIMContext ();
				
				SyncAndEmit ();
			} else {
				g_warning ("g_utf8_to_ucs4_fast failed for string '%s'", str);
			}
		}
	} else if (args->GetId () == TextBox::SelectionStartProperty) {
		length = abs (selection_cursor - selection_anchor);
		start = args->GetNewValue ()->AsInt32 ();
		
		if (start > buffer->len) {
			// clamp the selection start offset to a valid value
			SetSelectionStart (buffer->len);
			return;
		}
		
		if (start + length > buffer->len) {
			// clamp the selection length to a valid value
			BatchPush ();
			length = buffer->len - start;
			SetSelectionLength (length);
			BatchPop ();
		}
		
		// SelectionStartProperty is marked as AlwaysChange -
		// if the value hasn't actually changed, then we do
		// not want to emit the TextBoxModelChanged event.
		if (selection_anchor != start) {
			changed = TextBoxModelChangedSelection;
			have_offset = false;
		}
		
		// When set programatically, anchor is always the
		// start and cursor is always the end.
		selection_cursor = start + length;
		selection_anchor = start;
		
		emit |= SELECTION_CHANGED;
		
		SyncAndEmit ();
	} else if (args->GetId () == TextBox::SelectionLengthProperty) {
		start = MIN (selection_anchor, selection_cursor);
		length = args->GetNewValue ()->AsInt32 ();
		
		if (start + length > buffer->len) {
			// clamp the selection length to a valid value
			length = buffer->len - start;
			SetSelectionLength (length);
			return;
		}
		
		// SelectionLengthProperty is marked as AlwaysChange -
		// if the value hasn't actually changed, then we do
		// not want to emit the TextBoxModelChanged event.
		if (selection_cursor != start + length) {
			changed = TextBoxModelChangedSelection;
			have_offset = false;
		}
		
		// When set programatically, anchor is always the
		// start and cursor is always the end.
		selection_cursor = start + length;
		selection_anchor = start;
		
		emit |= SELECTION_CHANGED;
		
		SyncAndEmit ();
	} else if (args->GetId () == TextBox::SelectionBackgroundProperty) {
		changed = TextBoxModelChangedBrush;
	} else if (args->GetId () == TextBox::SelectionForegroundProperty) {
		changed = TextBoxModelChangedBrush;
	} else if (args->GetId () == TextBox::TextProperty) {
		if (setvalue) {
			Value *value = args->GetNewValue ();
			const char *str = value && value->AsString () ? value->AsString () : "";
			TextBoxUndoAction *action;
			gunichar *text;
			glong textlen;
			
			if ((text = g_utf8_to_ucs4_fast (str, -1, &textlen))) {
				if (buffer->len > 0) {
					// replace the current text
					action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, 0, buffer->len, text, textlen);
					
					buffer->Replace (0, buffer->len, text, textlen);
				} else {
					// insert the text
					action = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, 0, text, textlen);
					
					buffer->Insert (0, text, textlen);
				}
				
				undo->Push (action);
				redo->Clear ();
				
				emit |= TEXT_CHANGED;
				ClearSelection (0);
				ResetIMContext ();
				
				SyncAndEmit (value && !value->GetIsNull ());
			} else {
				g_warning ("g_utf8_to_ucs4_fast failed for string '%s'", str);
			}
		}
		
		changed = TextBoxModelChangedText;
	} else if (args->GetId () == TextBox::TextAlignmentProperty) {
		changed = TextBoxModelChangedTextAlignment;
	} else if (args->GetId () == TextBox::TextWrappingProperty) {
		changed = TextBoxModelChangedTextWrapping;
	} else if (args->GetId () == TextBox::HorizontalScrollBarVisibilityProperty) {
		// XXX more crap because these aren't templatebound.
		if (contentElement) {
			if ((prop = contentElement->GetDependencyProperty ("VerticalScrollBarVisibility")))
				contentElement->SetValue (prop, GetValue (TextBox::VerticalScrollBarVisibilityProperty));
		}
		
		Invalidate ();
	} else if (args->GetId () == TextBox::VerticalScrollBarVisibilityProperty) {
		// XXX more crap because these aren't templatebound.
		if (contentElement) {
			if ((prop = contentElement->GetDependencyProperty ("HorizontalScrollBarVisibility")))
				contentElement->SetValue (prop, GetValue (TextBox::HorizontalScrollBarVisibilityProperty));
		}
		
		Invalidate ();
	}
	
	if (changed != TextBoxModelChangedNothing)
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (changed, args));
	
	if (args->GetProperty ()->GetOwnerType () != Type::TEXTBOX) {
		TextBoxBase::OnPropertyChanged (args, error);
		return;
	}
	
	NotifyListenersOfPropertyChange (args, error);
}

void
TextBox::OnSubPropertyChanged (DependencyProperty *prop, DependencyObject *obj, PropertyChangedEventArgs *subobj_args)
{
	if (prop && (prop->GetId () == TextBox::SelectionBackgroundProperty ||
		     prop->GetId () == TextBox::SelectionForegroundProperty)) {
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (TextBoxModelChangedBrush));
		Invalidate ();
	}
	
	if (prop->GetOwnerType () != Type::TEXTBOX)
		TextBoxBase::OnSubPropertyChanged (prop, obj, subobj_args);
}

void
TextBox::OnApplyTemplate ()
{
	DependencyProperty *prop;
	
	TextBoxBase::OnApplyTemplate ();
	
	if (!contentElement)
		return;
	
	// XXX LAME these should be template bindings in the textbox template.
	if ((prop = contentElement->GetDependencyProperty ("VerticalScrollBarVisibility")))
		contentElement->SetValue (prop, GetValue (TextBox::VerticalScrollBarVisibilityProperty));
	
	if ((prop = contentElement->GetDependencyProperty ("HorizontalScrollBarVisibility")))
		contentElement->SetValue (prop, GetValue (TextBox::HorizontalScrollBarVisibilityProperty));
}


//
// PasswordBox
//

//
// PasswordBoxDynamicPropertyValueProvider
//

class PasswordBoxDynamicPropertyValueProvider : public PropertyValueProvider {
	Value *selection_background;
	Value *selection_foreground;
	
 public:
	PasswordBoxDynamicPropertyValueProvider (DependencyObject *obj, PropertyPrecedence precedence) : PropertyValueProvider (obj, precedence)
	{
		selection_background = NULL;
		selection_foreground = NULL;
	}
	
	virtual ~PasswordBoxDynamicPropertyValueProvider ()
	{
		delete selection_background;
		delete selection_foreground;
	}
	
	virtual Value *GetPropertyValue (DependencyProperty *property)
	{
		if (property->GetId () == PasswordBox::SelectionBackgroundProperty) {
			return selection_background;
		} else if (property->GetId () == PasswordBox::SelectionForegroundProperty) {
			return selection_foreground;
		}
		
		return NULL;
	}
	
	void InitializeSelectionBrushes ()
	{
		if (!selection_background)
			selection_background = Value::CreateUnrefPtr (new SolidColorBrush ("#FF444444"));
		
		if (!selection_foreground)
			selection_foreground = Value::CreateUnrefPtr (new SolidColorBrush ("#FFFFFFFF"));
	}
};


//
// PasswordBox
//

PasswordBox::PasswordBox ()
{
	providers[PropertyPrecedence_DynamicValue] = new PasswordBoxDynamicPropertyValueProvider (this, PropertyPrecedence_DynamicValue);
	
	Initialize (Type::PASSWORDBOX, "System.Windows.Controls.PasswordBox");
	events_mask = TEXT_CHANGED;
	
	display = g_string_new ("");
}

PasswordBox::~PasswordBox ()
{
	g_string_free (display, true);
}

int
PasswordBox::CursorDown (int cursor, bool page)
{
	return GetBuffer ()->len;
}

int
PasswordBox::CursorUp (int cursor, bool page)
{
	return 0;
}

int
PasswordBox::CursorLineBegin (int cursor)
{
	return 0;
}

int
PasswordBox::CursorLineEnd (int cursor, bool include)
{
	return GetBuffer ()->len;
}

int
PasswordBox::CursorNextWord (int cursor)
{
	return GetBuffer ()->len;
}

int
PasswordBox::CursorPrevWord (int cursor)
{
	return 0;
}

void
PasswordBox::EmitTextChanged ()
{
	Emit (PasswordBox::PasswordChangedEvent, new RoutedEventArgs ());
}

void
PasswordBox::SyncSelectedText ()
{
	if (selection_cursor != selection_anchor) {
		int length = abs (selection_cursor - selection_anchor);
		int start = MIN (selection_anchor, selection_cursor);
		char *text;
		
		text = g_ucs4_to_utf8 (buffer->text + start, length, NULL, NULL, NULL);
		
		setvalue = false;
		SetValue (PasswordBox::SelectedTextProperty, Value (text, true));
		setvalue = true;
	} else {
		setvalue = false;
		SetValue (PasswordBox::SelectedTextProperty, Value (""));
		setvalue = true;
	}
}

void
PasswordBox::SyncDisplayText ()
{
	gunichar c = GetPasswordChar ();
	
	g_string_truncate (display, 0);
	
	for (int i = 0; i < buffer->len; i++)
		g_string_append_unichar (display, c);
}

void
PasswordBox::SyncText ()
{
	char *text = g_ucs4_to_utf8 (buffer->text, buffer->len, NULL, NULL, NULL);
	
	setvalue = false;
	SetValue (PasswordBox::PasswordProperty, Value (text, true));
	setvalue = true;
	
	SyncDisplayText ();
}

const char *
PasswordBox::GetDisplayText ()
{
	return display->str;
}

void
PasswordBox::ClearFontSource ()
{
	ClearValue (PasswordBox::FontSourceProperty);
}

void
PasswordBox::OnPropertyChanged (PropertyChangedEventArgs *args, MoonError *error)
{
	TextBoxModelChangeType changed = TextBoxModelChangedNothing;
	int length, start;
	
	if (args->GetId () == PasswordBox::MaxLengthProperty) {
		// update max_length state
		max_length = args->GetNewValue()->AsInt32 ();
	} else if (args->GetId () == PasswordBox::PasswordCharProperty) {
		changed = TextBoxModelChangedText;
	} else if (args->GetId () == PasswordBox::PasswordProperty) {
		if (setvalue) {
			Value *value = args->GetNewValue ();
			const char *str = value && value->AsString () ? value->AsString () : "";
			TextBoxUndoAction *action;
			gunichar *text;
			glong textlen;
			
			if ((text = g_utf8_to_ucs4_fast (str, -1, &textlen))) {
				if (buffer->len > 0) {
					// replace the current text
					action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, 0, buffer->len, text, textlen);
					
					buffer->Replace (0, buffer->len, text, textlen);
				} else {
					// insert the text
					action = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, 0, text, textlen);
					
					buffer->Insert (0, text, textlen);
				}
				
				undo->Push (action);
				redo->Clear ();
				
				emit |= TEXT_CHANGED;
				SyncDisplayText ();
				ClearSelection (0);
				ResetIMContext ();
				
				SyncAndEmit ();
			}
		}
		
		changed = TextBoxModelChangedText;
	} else if (args->GetId () == PasswordBox::SelectedTextProperty) {
		if (setvalue) {
			Value *value = args->GetNewValue ();
			const char *str = value && value->AsString () ? value->AsString () : "";
			TextBoxUndoAction *action;
			gunichar *text;
			glong textlen;
			
			length = abs (selection_cursor - selection_anchor);
			start = MIN (selection_anchor, selection_cursor);
			
			if ((text = g_utf8_to_ucs4_fast (str, -1, &textlen))) {
				if (length > 0) {
					// replace the currently selected text
					action = new TextBoxUndoActionReplace (selection_anchor, selection_cursor, buffer, start, length, text, textlen);
					
					buffer->Replace (start, length, text, textlen);
				} else {
					// insert the text at the cursor
					action = new TextBoxUndoActionInsert (selection_anchor, selection_cursor, start, text, textlen);
					
					buffer->Insert (start, text, textlen);
				}
				
				undo->Push (action);
				redo->Clear ();
				
				ClearSelection (start + textlen);
				emit |= TEXT_CHANGED;
				SyncDisplayText ();
				ResetIMContext ();
				
				SyncAndEmit ();
			}
		}
	} else if (args->GetId () == PasswordBox::SelectionStartProperty) {
		length = abs (selection_cursor - selection_anchor);
		start = args->GetNewValue ()->AsInt32 ();
		
		if (start > buffer->len) {
			// clamp the selection start offset to a valid value
			SetSelectionStart (buffer->len);
			return;
		}
		
		if (start + length > buffer->len) {
			// clamp the selection length to a valid value
			BatchPush ();
			length = buffer->len - start;
			SetSelectionLength (length);
			BatchPop ();
		}
		
		// SelectionStartProperty is marked as AlwaysChange -
		// if the value hasn't actually changed, then we do
		// not want to emit the TextBoxModelChanged event.
		if (selection_anchor != start) {
			changed = TextBoxModelChangedSelection;
			have_offset = false;
		}
		
		// When set programatically, anchor is always the
		// start and cursor is always the end.
		selection_cursor = start + length;
		selection_anchor = start;
		
		emit |= SELECTION_CHANGED;
		
		SyncAndEmit ();
	} else if (args->GetId () == PasswordBox::SelectionLengthProperty) {
		start = MIN (selection_anchor, selection_cursor);
		length = args->GetNewValue ()->AsInt32 ();
		
		if (start + length > buffer->len) {
			// clamp the selection length to a valid value
			length = buffer->len - start;
			SetSelectionLength (length);
			return;
		}
		
		// SelectionLengthProperty is marked as AlwaysChange -
		// if the value hasn't actually changed, then we do
		// not want to emit the TextBoxModelChanged event.
		if (selection_cursor != start + length) {
			changed = TextBoxModelChangedSelection;
			have_offset = false;
		}
		
		// When set programatically, anchor is always the
		// start and cursor is always the end.
		selection_cursor = start + length;
		selection_anchor = start;
		
		emit |= SELECTION_CHANGED;
		
		SyncAndEmit ();
	} else if (args->GetId () == PasswordBox::SelectionBackgroundProperty) {
		changed = TextBoxModelChangedBrush;
	} else if (args->GetId () == PasswordBox::SelectionForegroundProperty) {
		changed = TextBoxModelChangedBrush;
	}
	
	if (changed != TextBoxModelChangedNothing)
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (changed, args));
	
	if (args->GetProperty ()->GetOwnerType () != Type::TEXTBOX) {
		TextBoxBase::OnPropertyChanged (args, error);
		return;
	}
	
	NotifyListenersOfPropertyChange (args, error);
}

void
PasswordBox::OnSubPropertyChanged (DependencyProperty *prop, DependencyObject *obj, PropertyChangedEventArgs *subobj_args)
{
	if (prop && (prop->GetId () == PasswordBox::SelectionBackgroundProperty ||
		     prop->GetId () == PasswordBox::SelectionForegroundProperty)) {
		Emit (ModelChangedEvent, new TextBoxModelChangedEventArgs (TextBoxModelChangedBrush));
		Invalidate ();
	}
	
	if (prop->GetOwnerType () != Type::TEXTBOX)
		TextBoxBase::OnSubPropertyChanged (prop, obj, subobj_args);
}


//
// TextBoxView
//

#define CURSOR_BLINK_TIMEOUT_DEFAULT  900
#define CURSOR_BLINK_ON_MULTIPLIER    2
#define CURSOR_BLINK_OFF_MULTIPLIER   1
#define CURSOR_BLINK_DELAY_MULTIPLIER 3
#define CURSOR_BLINK_DIVIDER          3

TextBoxView::TextBoxView ()
{
	SetObjectType (Type::TEXTBOXVIEW);
	
	AddHandler (UIElement::MouseLeftButtonDownEvent, TextBoxView::mouse_left_button_down, this);
	AddHandler (UIElement::MouseLeftButtonUpEvent, TextBoxView::mouse_left_button_up, this);
	
	SetCursor (MouseCursorIBeam);
	
	cursor = Rect (0, 0, 0, 0);
	layout = new TextLayout ();
	selection_changed = false;
	had_selected_text = false;
	cursor_visible = false;
	blink_timeout = 0;
	textbox = NULL;
	dirty = false;
}

TextBoxView::~TextBoxView ()
{
	RemoveHandler (UIElement::MouseLeftButtonDownEvent, TextBoxView::mouse_left_button_down, this);
	RemoveHandler (UIElement::MouseLeftButtonUpEvent, TextBoxView::mouse_left_button_up, this);
	
	if (textbox) {
		textbox->RemoveHandler (TextBox::ModelChangedEvent, TextBoxView::model_changed, this);
		textbox->view = NULL;
	}
	
	DisconnectBlinkTimeout ();
	
	delete layout;
}

TextLayoutLine *
TextBoxView::GetLineFromY (double y, int *index)
{
	return layout->GetLineFromY (Point (), y, index);
}

TextLayoutLine *
TextBoxView::GetLineFromIndex (int index)
{
	return layout->GetLineFromIndex (index);
}

int
TextBoxView::GetCursorFromXY (double x, double y)
{
	return layout->GetCursorFromXY (Point (), x, y);
}

gboolean
TextBoxView::blink (void *user_data)
{
	return ((TextBoxView *) user_data)->Blink ();
}

static guint
GetCursorBlinkTimeout (TextBoxView *view)
{
	GtkSettings *settings;
	MoonWindow *window;
	GdkScreen *screen;
	GdkWindow *widget;
	Surface *surface;
	guint timeout;
	
	if (!(surface = view->GetSurface ()))
		return CURSOR_BLINK_TIMEOUT_DEFAULT;
	
	if (!(window = surface->GetWindow ()))
		return CURSOR_BLINK_TIMEOUT_DEFAULT;
	
	if (!(widget = window->GetGdkWindow ()))
		return CURSOR_BLINK_TIMEOUT_DEFAULT;
	
	if (!(screen = gdk_drawable_get_screen ((GdkDrawable *) widget)))
		return CURSOR_BLINK_TIMEOUT_DEFAULT;
	
	if (!(settings = gtk_settings_get_for_screen (screen)))
		return CURSOR_BLINK_TIMEOUT_DEFAULT;
	
	g_object_get (settings, "gtk-cursor-blink-time", &timeout, NULL);
	
	return timeout;
}

void
TextBoxView::ConnectBlinkTimeout (guint multiplier)
{
	guint timeout = GetCursorBlinkTimeout (this) * multiplier / CURSOR_BLINK_DIVIDER;
	Surface *surface = GetSurface ();
	TimeManager *manager;
	
	if (!surface || !(manager = surface->GetTimeManager ()))
		return;
	
	blink_timeout = manager->AddTimeout (MOON_PRIORITY_DEFAULT, timeout, TextBoxView::blink, this);
}

void
TextBoxView::DisconnectBlinkTimeout ()
{
	TimeManager *manager;
	Surface *surface;
	
	if (blink_timeout != 0) {
		if (!(surface = GetSurface ()) || !(manager = surface->GetTimeManager ()))
			return;
		
		manager->RemoveTimeout (blink_timeout);
		blink_timeout = 0;
	}
}

bool
TextBoxView::Blink ()
{
	guint multiplier;
	
	Deployment::SetCurrent (GetDeployment ());
	
	if (cursor_visible) {
		multiplier = CURSOR_BLINK_OFF_MULTIPLIER;
		HideCursor ();
	} else {
		multiplier = CURSOR_BLINK_ON_MULTIPLIER;
		ShowCursor ();
	}
	
	ConnectBlinkTimeout (multiplier);
	
	return false;
}

void
TextBoxView::DelayCursorBlink ()
{
	DisconnectBlinkTimeout ();
	ConnectBlinkTimeout (CURSOR_BLINK_DELAY_MULTIPLIER);
	UpdateCursor (true);
	ShowCursor ();
}

void
TextBoxView::BeginCursorBlink ()
{
	if (blink_timeout == 0) {
		ConnectBlinkTimeout (CURSOR_BLINK_ON_MULTIPLIER);
		UpdateCursor (true);
		ShowCursor ();
	}
}

void
TextBoxView::EndCursorBlink ()
{
	DisconnectBlinkTimeout ();
	
	if (cursor_visible)
		HideCursor ();
}

void
TextBoxView::ResetCursorBlink (bool delay)
{
	if (textbox->IsFocused () && !textbox->HasSelectedText ()) {
		// cursor is blinkable... proceed with blinkage
		if (delay)
			DelayCursorBlink ();
		else
			BeginCursorBlink ();
	} else {
		// cursor not blinkable... stop all blinkage
		EndCursorBlink ();
	}
}

void
TextBoxView::InvalidateCursor ()
{
	Invalidate (cursor.Transform (&absolute_xform));
}

void
TextBoxView::ShowCursor ()
{
	cursor_visible = true;
	InvalidateCursor ();
}

void
TextBoxView::HideCursor ()
{
	cursor_visible = false;
	InvalidateCursor ();
}

void
TextBoxView::UpdateCursor (bool invalidate)
{
	int cur = textbox->GetCursor ();
	GdkRectangle area;
	Rect rect;
	
	// invalidate current cursor rect
	if (invalidate && cursor_visible)
		InvalidateCursor ();
	
	// calculate the new cursor rect
	cursor = layout->GetCursor (Point (), cur);
	
	// transform the cursor rect into absolute coordinates for the IM context
	rect = cursor.Transform (&absolute_xform);
	area = rect.ToGdkRectangle ();
	
	gtk_im_context_set_cursor_location (textbox->im_ctx, &area);
	
	textbox->EmitCursorPositionChanged (cursor.height, cursor.x, cursor.y);
	
	// invalidate the new cursor rect
	if (invalidate && cursor_visible)
		InvalidateCursor ();
}

void
TextBoxView::UpdateText ()
{
	const char *text = textbox->GetDisplayText ();
	
	layout->SetText (text ? text : "", -1);
}

void
TextBoxView::GetSizeForBrush (cairo_t *cr, double *width, double *height)
{
	*height = GetActualHeight ();
	*width = GetActualWidth ();
}

Size
TextBoxView::MeasureOverride (Size availableSize)
{
	Size desired = Size ();
	
	Layout (availableSize);
	
	layout->GetActualExtents (&desired.width, &desired.height);
	
	return desired.Min (availableSize);
}

Size
TextBoxView::ArrangeOverride (Size finalSize)
{
	Size arranged = Size ();
	
	Layout (finalSize);
	
	layout->GetActualExtents (&arranged.width, &arranged.height);
	
	return arranged.Max (finalSize);
}

void
TextBoxView::Layout (Size constraint)
{
	layout->SetMaxWidth (constraint.width);
	
	layout->Layout ();
}

void
TextBoxView::Paint (cairo_t *cr)
{
	layout->Render (cr, GetOriginPoint (), Point ());
	
	if (cursor_visible) {
		cairo_antialias_t alias = cairo_get_antialias (cr);
		double h = round (cursor.height);
		double x = cursor.x;
		double y = cursor.y;
		
		// disable antialiasing
		cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
		
		// set the color to black
		cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
		
		// snap 'x' to the half-pixel grid (to try and get a sharp 1-pixel-wide line)
		cairo_user_to_device (cr, &x, &y);
		x = trunc (x) + 0.5; y = trunc (y);
		cairo_device_to_user (cr, &x, &y);
		
		// draw the cursor
		cairo_set_line_width (cr, 1.0);
		cairo_move_to (cr, x, y);
		cairo_line_to (cr, x, y + h);
		cairo_stroke (cr);
		
		// restore antialiasing
		cairo_set_antialias (cr, alias);
	}
}

void
TextBoxView::Render (cairo_t *cr, Region *region, bool path_only)
{
	TextBoxDynamicPropertyValueProvider *dynamic = (TextBoxDynamicPropertyValueProvider *) textbox->providers[PropertyPrecedence_DynamicValue];
	Size renderSize = GetRenderSize ();
	
	dynamic->InitializeSelectionBrushes ();
	
	if (dirty) {
		Layout (renderSize);
		UpdateCursor (false);
		dirty = false;
	}
	
	if (selection_changed) {
		layout->Select (textbox->GetSelectionStart (), textbox->GetSelectionLength ());
		selection_changed = false;
	}
	
	cairo_save (cr);
	cairo_set_matrix (cr, &absolute_xform);
	layout->SetAvailableWidth (renderSize.width);
	Paint (cr);
	cairo_restore (cr);
}

void
TextBoxView::OnModelChanged (TextBoxModelChangedEventArgs *args)
{
	switch (args->changed) {
	case TextBoxModelChangedTextAlignment:
		// text alignment changed, update our layout
		if (layout->SetTextAlignment ((TextAlignment) args->property->GetNewValue()->AsInt32 ()))
			dirty = true;
		break;
	case TextBoxModelChangedTextWrapping:
		// text wrapping changed, update our layout
		if (layout->SetTextWrapping ((TextWrapping) args->property->GetNewValue()->AsInt32 ()))
			dirty = true;
		break;
	case TextBoxModelChangedSelection:
		if (had_selected_text || textbox->HasSelectedText ()) {
			// the selection has changed, update the layout's selection
			had_selected_text = textbox->HasSelectedText ();
			selection_changed = true;
			ResetCursorBlink (false);
		} else {
			// cursor position changed
			ResetCursorBlink (true);
			return;
		}
		break;
	case TextBoxModelChangedBrush:
		// a brush has changed, no layout updates needed, we just need to re-render
		break;
	case TextBoxModelChangedFont:
		// font changed, need to recalculate layout/bounds
		layout->ResetState ();
		dirty = true;
		break;
	case TextBoxModelChangedText:
		// the text has changed, need to recalculate layout/bounds
		UpdateText ();
		dirty = true;
		break;
	default:
		// nothing changed??
		return;
	}
	
	if (dirty)
		UpdateBounds (true);
	
	Invalidate ();
}

void
TextBoxView::model_changed (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxView *) closure)->OnModelChanged ((TextBoxModelChangedEventArgs *) args);
}

void
TextBoxView::OnFocusOut ()
{
	EndCursorBlink ();
}

void
TextBoxView::OnFocusIn ()
{
	ResetCursorBlink (false);
}

void
TextBoxView::OnMouseLeftButtonDown (MouseEventArgs *args)
{
	// proxy to our parent TextBox control
	textbox->OnMouseLeftButtonDown (args);
}

void
TextBoxView::mouse_left_button_down (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxView *) closure)->OnMouseLeftButtonDown ((MouseEventArgs *) args);
}

void
TextBoxView::OnMouseLeftButtonUp (MouseEventArgs *args)
{
	// proxy to our parent TextBox control
	textbox->OnMouseLeftButtonUp (args);
}

void
TextBoxView::mouse_left_button_up (EventObject *sender, EventArgs *args, gpointer closure)
{
	((TextBoxView *) closure)->OnMouseLeftButtonUp ((MouseEventArgs *) args);
}

void
TextBoxView::SetTextBox (TextBoxBase *textbox)
{
	TextLayoutAttributes *attrs;
	
	if (this->textbox == textbox)
		return;
	
	if (this->textbox) {
		// remove the event handlers from the old textbox
		this->textbox->RemoveHandler (TextBoxBase::ModelChangedEvent, TextBoxView::model_changed, this);
	}
	
	this->textbox = textbox;
	
	if (textbox) {
		textbox->AddHandler (TextBoxBase::ModelChangedEvent, TextBoxView::model_changed, this);
		
		// sync our state with the textbox
		layout->SetTextAttributes (new List ());
		attrs = new TextLayoutAttributes ((ITextAttributes *) textbox, 0);
		layout->GetTextAttributes ()->Append (attrs);
		
		layout->SetTextAlignment (textbox->GetTextAlignment ());
		layout->SetTextWrapping (textbox->GetTextWrapping ());
		had_selected_text = textbox->HasSelectedText ();
		selection_changed = true;
		UpdateText ();
	} else {
		layout->SetTextAttributes (NULL);
		layout->SetText (NULL, -1);
	}
	
	UpdateBounds (true);
	Invalidate ();
	dirty = true;
}
